/*
 *  Copyright (C) 2006 Qian Shen <qshen@redhat.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define Uses_SCIM_HELPER
#define Uses_SCIM_CONFIG_BASE
#include <map>
#include <scim.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include "scim_pinyin_helper.h"

using namespace scim;

#define scim_module_init pinyin_imengine_helper_LTX_scim_module_init
#define scim_module_exit pinyin_imengine_helper_LTX_scim_module_exit
#define scim_helper_module_number_of_helpers pinyin_imengine_helper_LTX_scim_helper_module_number_of_helpers
#define scim_helper_module_get_helper_info pinyin_imengine_helper_LTX_scim_helper_module_get_helper_info
#define scim_helper_module_run_helper pinyin_imengine_helper_LTX_scim_helper_module_run_helper

static void       run                         (const String        &display,
                                               const ConfigPointer &config);
					       
HelperAgent helper_agent;					       

HelperInfo helper_info (SCIM_PINYIN_HELPER_UUID,        // uuid
                        "",                            // name
                        "",                            // icon
                        "",
                        SCIM_HELPER_NEED_SCREEN_INFO|SCIM_HELPER_NEED_SCREEN_INFO);
			
//Module Interface
extern "C" {
    void scim_module_init (void)
    {
        bindtextdomain (GETTEXT_PACKAGE, SCIM_PINYIN_LOCALEDIR);
        bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");

        helper_info.name = String (_("Smart pinyin helper"));
        helper_info.description = String (
            _("A helper module for Smart Pinyin IMEngine."));
    }

    void scim_module_exit (void)
    {
    }

    unsigned int scim_helper_module_number_of_helpers (void)
    {
        return 1;
    }

    bool scim_helper_module_get_helper_info (unsigned int idx, HelperInfo &info)
    {
        if (idx == 0) {
            info = helper_info; 
            return true;
        }
        return false;
    }

    void scim_helper_module_run_helper (const String &uuid,
                                        const ConfigPointer &config,
                                        const String &display)
    {
        SCIM_DEBUG_MAIN(1) << "scim_pinyin_imengine_helper_LTX_scim_helper_module_run_helper ()\n";

        if (uuid == String (SCIM_PINYIN_HELPER_UUID)) {
            run (display, config);
        }

        SCIM_DEBUG_MAIN(1) << "exit scim_pinyin_imengine_helper_LTX_scim_helper_module_run_helper ()\n";
    }
}

//Global Variables
static int        cur_ic = -1;
static String     cur_uuid;

GtkWidget    *	  BaseWindow;
GtkListStore *	  list_store;
GtkTreeModel *    tree_model;
GtkTreeView  *	  tree_view;
GtkTreeIter 	  cur_iter;
GtkCellRenderer * renderer_num, *renderer_phrase, *renderer_pinyin;
GtkWidget    *	  submit_button, *disable_button, *add_button;
GtkWidget    *    add_phrase_row, *disable_phrase_row; 

uint32		  max_user_phrase_length = 8;

bool		  ok_to_add_new_row = false;
guint		  row_count = 0;

int		  list_cur_num = -1;

static gboolean
helper_agent_input_handler (GIOChannel *source,
                            GIOCondition condition,
                            gpointer user_data)
{
    if (condition == G_IO_IN) {
        HelperAgent *agent = static_cast<HelperAgent*> (user_data);
        if (agent && agent->has_pending_event ())
            agent->filter_event ();
    } else if (condition == G_IO_ERR || condition == G_IO_HUP) {
        gtk_main_quit ();
    }
    return TRUE;
}

static void
slot_exit (const HelperAgent *agent, int ic, const String &uuid)
{
    gtk_main_quit ();
}

static void slot_attach_input_context (const HelperAgent *agent, int ic, const String &uuid)
{
    std::cout << "slot_attach_input_context (" << ic << ", " << uuid << ")\n";

    cur_ic = ic;
    cur_uuid = uuid;

    // Inform the scim-pinyin imengine that scim-pinyin-helper started successfully
    Transaction trans;
    trans.put_command (SCIM_TRANS_CMD_REQUEST);
    trans.put_command (SCIM_TRANS_CMD_START_HELPER);
    agent->send_imengine_event (ic, uuid, trans);
}

static void slot_detach_input_context (const HelperAgent *agent, int ic, const String &uuid)
{
    std::cout << "slot_detach_input_context (" << ic << ", " << uuid << ")\n";
}

static void slot_update_screen (const HelperAgent *, int ic, const String &uuid, int screen)
{
    std::cout << "slot_update_screen (" << ic << ", " << uuid << ", " << screen << ")\n";
}

static void slot_update_spot_location (const HelperAgent *, int ic, const String &uuid, int x, int y)
{
    std::cout << "slot_update_spot_location (" << ic << ", " << uuid << ", " << x << ", " << y << ")\n";

    //gtk_window_move (GTK_WINDOW (main_window), x + 16, y + 16);
}

void
on_button_quit_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_main_quit ();   
}

void
on_quit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    gtk_main_quit ();
}

void
on_help_online_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    char *help_text = _("SCIM Pinyin Helper\n\n"
        "Copyright (C) 2006 Qian Shen <qshen@redhat.com>\n\n"
        "SCIM Pinyin Helper is a utility for users to manage their preferred phrase in Pinyin \n"
	"input method.\n\n\n"
	"Command Options:\n\n"
	"  Show User Phrase Library: lists current users' phrase library. \n\n"
	"  Import User Phrase Library: imports user's phrase library file. \n\n"
	"  Export User Phrase Library: exports user's phrase library file. \n\n"
	"  Add New Phrases: adds new user phrases into current scim pinyin user phrase library. \n\n"
	"  Delete Phrase: removes phrases from current scim pinyin user phrase library. \n\n\n"
	"Notes: \n"
	"  User is not recommended to modify the exported phrase library file. If user insisted \n"
	"  to insert new phrases manually, the phrase name (column 1) has to be matching the \n"
	"  pinyin name (column 2) located in pinyin table. Otherwise, certain phrases might not \n"
	"  able to be imported successfully\n\n");

    // Create help window - copy from "scim_panel_gtk.cpp"
    {
    
	GtkWidget *_help_dialog, *_help_scroll, *_help_area, *_command_menu;    
        GtkWidget *frame;
        GtkWidget *vbox;
        GtkWidget *scroll;

        _help_dialog = gtk_dialog_new_with_buttons (_("SCIM Help"),
                                NULL,
                                GtkDialogFlags (0),
                                GTK_STOCK_OK,
                                GTK_RESPONSE_OK,
                                NULL);

        g_signal_connect_swapped (GTK_OBJECT (_help_dialog), 
                                  "response", 
                                  G_CALLBACK (gtk_widget_hide),
                                  GTK_OBJECT (_help_dialog));

        g_signal_connect_swapped (GTK_OBJECT (_help_dialog), 
                                  "delete_event", 
                                  G_CALLBACK (gtk_widget_hide_on_delete),
                                  GTK_OBJECT (_help_dialog));

        frame = gtk_frame_new (_("Smart Common Input Method"));

        gtk_box_pack_start (GTK_BOX (GTK_DIALOG (_help_dialog)->vbox), frame, TRUE, TRUE, 0);
        gtk_widget_show (frame);

        vbox = gtk_vbox_new (FALSE, 8);
        gtk_container_add (GTK_CONTAINER (frame), vbox);
        gtk_widget_show (vbox);

        _help_scroll = gtk_scrolled_window_new (NULL, NULL);
        gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (_help_scroll), GTK_POLICY_NEVER, GTK_POLICY_NEVER);
        gtk_box_pack_start (GTK_BOX (vbox), _help_scroll, TRUE, TRUE, 0);
        gtk_widget_show (_help_scroll);

        _help_area = gtk_label_new (help_text);
        gtk_label_set_justify (GTK_LABEL (_help_area), GTK_JUSTIFY_LEFT);
        gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (_help_scroll), _help_area);
        gtk_widget_show (_help_area);
	
	
	gtk_dialog_run( GTK_DIALOG(_help_dialog) );
	
	gtk_widget_destroy(_help_dialog); 	
    }

}

void
on_new_row_activate                     (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
{
    std::cout << "in on_new_column_activate()\n";
    
    if( !ok_to_add_new_row )
        return;
   
    GtkTreeIter iter;
    GtkTreePath *path; 
    GtkTreeViewColumn* column;    
     
    gtk_list_store_append (list_store, &iter);
   
    row_count ++;
    
    gtk_list_store_set (list_store, &iter,
			  NUM_COLUMN, row_count,
			  PHRASE_COLUMN, "",
			  PINYIN_COLUMN, "",
			  -1);

    path = gtk_tree_model_get_path ( tree_model, &iter );
    
    column = gtk_tree_view_get_column ( tree_view, 1 );
    	
    gtk_tree_view_set_cursor ( tree_view,
                               path,
                               column,
                               TRUE);  
    // bug? - if use this api to grab focus, scim will crash!			       
    //gtk_widget_grab_focus ( (GtkWidget*)tree_view);
    
    ok_to_add_new_row = false;			         			 
}

void
on_list_user_phrase_library             (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
{
    gtk_widget_hide (submit_button);
    gtk_widget_show (disable_button);
    gtk_widget_show (add_button); 
    gtk_widget_show (add_phrase_row);
    gtk_widget_show (disable_phrase_row);     

    g_object_set ( renderer_phrase, "editable", false, NULL);
    g_object_set ( renderer_pinyin, "editable", false, NULL);   

    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB);
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
    std::cout << "in on_list_user_phrase_library(), send : SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB\n";
}

void 
on_dir_select_ok			(GtkWidget *widget, 
					 gpointer user_data) 
{
    GtkWidget 	*dir_selector = GTK_WIDGET (user_data);
    const gchar *selected_dirname;
    String   	string_dirname;
    
    selected_dirname = gtk_file_selection_get_filename (GTK_FILE_SELECTION (dir_selector));
    string_dirname = String( selected_dirname );
   
    if( g_file_test( selected_dirname , G_FILE_TEST_IS_DIR ) )
        gtk_widget_destroy( GTK_WIDGET(dir_selector) );
    else
	return;  
	
    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_SAVE_USER_LIB);
    send.put_data ( string_dirname );
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
    std::cout << "in on_dir_select_ok(), send : SCIM_PINYIN_TRANS_CMD_SAVE_USER_LIB\n";      
}					 

/* save the 3 user phrase libraries into a directory
   this function should be modified into 'bakup library'
*/
void
on_save_user_phrase_library             (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
{
   GtkWidget *dir_selector;
  
   dir_selector = gtk_file_selection_new (_("please select a directory."));
   
   g_signal_connect (GTK_FILE_SELECTION (dir_selector)->ok_button,
                     "clicked",
                     G_CALLBACK (on_dir_select_ok),
                     dir_selector);
   			   
   g_signal_connect_swapped (GTK_FILE_SELECTION (dir_selector)->cancel_button,
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy),
                             dir_selector); 
   
   gtk_widget_show (dir_selector);   
}

void 
on_lib_select_ok			(GtkWidget *widget, 
					 gpointer user_data) 
{
   GtkWidget *lib_selector = GTK_WIDGET (user_data);
   const gchar *selected_dirname;
   String string_dirname;
   
   String pinyin_lib_bakname("/phrase_lib.bak");   
   String pinyin_phrase_lib_bakname("/pinyin_phrase_lib.bak");
   String pinyin_phrase_index_bakname("/pinyin_phrase_index.bak");    

   selected_dirname = gtk_file_selection_get_filename (GTK_FILE_SELECTION (lib_selector));
   string_dirname = String( selected_dirname );
   
   FILE *user_lib;
   gchar *path;
   
   
   /* test the libraries --by name */
   if( g_file_test( selected_dirname , G_FILE_TEST_IS_DIR ) )
   {
     std::cout << "in on_lib_select_ok(), send : selected_dirname=" << selected_dirname << "\n"; 
     
     if( g_file_test( (string_dirname + pinyin_lib_bakname).c_str() , G_FILE_TEST_EXISTS) &&
         g_file_test( (string_dirname + pinyin_phrase_lib_bakname).c_str() , G_FILE_TEST_EXISTS) &&
	 g_file_test( (string_dirname + pinyin_phrase_index_bakname).c_str() , G_FILE_TEST_EXISTS) ) 
     {     
         gtk_widget_destroy( GTK_WIDGET(lib_selector) );
     }
     else
     {
         GtkWidget *dialog;
	 dialog = gtk_message_dialog_new( GTK_WINDOW(lib_selector), 
	    				     GTK_DIALOG_MODAL,
					     GTK_MESSAGE_ERROR,
					     GTK_BUTTONS_OK,
					     _("Invalid phrase libraries in this directory!") );
	 gtk_dialog_run( GTK_DIALOG(dialog) );
	 gtk_widget_destroy(dialog);  
	    
	 return;
      }       
   }
   else
	return;  


    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB);
    send.put_command (SCIM_PINYIN_TRANS_CMD_COMMIT_NEW_PHRASE);
    send.put_data ( string_dirname );
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
    std::cout << "in on_dir_select_ok(), send : SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB\n"; 
}

/* To copy the 3 phrase libraries saved in a directory before into the scim-pinyin user lib dir.
   will change it to called 'resume library'
*/
void
on_import_user_phrase_library_old       (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
{
   GtkWidget *lib_selector;
 
   lib_selector = gtk_file_selection_new (_("please select the directory that contains the user phrase library."));
   
   g_signal_connect (GTK_FILE_SELECTION (lib_selector)->ok_button,
                     "clicked",
                     G_CALLBACK (on_lib_select_ok),
                     lib_selector);
   
   g_signal_connect_swapped (GTK_FILE_SELECTION (lib_selector)->cancel_button,
                             "clicked",
                             G_CALLBACK (gtk_widget_destroy),
                             lib_selector); 
   
   gtk_widget_show (lib_selector);
}

void
on_import_user_phrase_library           (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
{
    GtkWidget * dialog;
 
    dialog = gtk_file_chooser_dialog_new (_("please select the user phrase library saved before."),
				          NULL,
				          GTK_FILE_CHOOSER_ACTION_OPEN,
				          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				          GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
				          NULL);
					  
    if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
        char    *filename, buf[100];
	FILE 	*user_lib;
        WideString  w_phrase;
	WideString  w_pinyin;
        std::vector <WideString>    phrase_vec;
        std::vector <WideString>    pinyin_vec;	
	
	/* open the user libary file */
        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));    
	
    	user_lib = fopen( filename, "r");
	
    	if( !user_lib )
    	{
             GtkWidget *dialog;
	     dialog = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL,
					     GTK_MESSAGE_ERROR,
					     GTK_BUTTONS_OK,
					     _("Faild to open pinyin user phrase library!") );
	     gtk_dialog_run( GTK_DIALOG(dialog) );
	     gtk_widget_destroy(dialog);
	      
	     fclose(user_lib); 	    
	     return;
         } 
	
	/* parse the user library file, 
	   store pinyin strings into pinyin_vec, 
	   store phrase strings into phrase_vec.
	*/
	
        pinyin_vec.clear();
        phrase_vec.clear();	
	    
        while( fgets(buf, 100, user_lib ) != NULL ){
            if( buf[0] == '#' )
	        continue;
		
	    int i;
            int seperator = -1;//record the position of the first space between phrase and pinyin
	    int endnum = -1;   //recore the position of the last word in the line --'\n'
	
	    for( i = 0; buf[i] != '\n' && i <100; i++){
	        if( buf[i] == ' ')
	            if( buf[i+1] == ' ' &&  buf[i+2] == ' ' &&  buf[i+3] == ' ')
		        seperator = i;
	    }
	    if ( buf[i] == '\n')	
	        endnum = i;
	
	    if( seperator > 0 && endnum > 0){
	        gchar phrase [seperator ];
	        gchar pinyin [endnum - seperator -4 ];
	    
	        for ( i=0; i < seperator; i++) 
	            phrase[i] = buf[i];
	        phrase[i]='\0';
	    
	        for ( i=0; i < endnum - seperator -4 -1; i++) 
	            pinyin[i] = buf[seperator+4+i];
	        pinyin[i]='\0';
	    
	        std::cout << phrase << "    "  << pinyin <<"\n";
	    
	        w_phrase = utf8_mbstowcs( phrase );
	        w_pinyin = utf8_mbstowcs( pinyin );	    	
	    
	        pinyin_vec.push_back( w_pinyin );
	        phrase_vec.push_back( w_phrase );
	    }

        }
	
        fclose(user_lib);
	        
        Transaction send;
        send.put_command (SCIM_TRANS_CMD_REQUEST);
        send.put_command (SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB);
        send.put_data ( phrase_vec );
        send.put_data ( pinyin_vec );
        helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
        std::cout << "in on_dir_select_ok(), send : SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB\n"; 		  	
    }
    
    gtk_widget_destroy (dialog); 					  					  
}

void
on_export_user_phrase_library           (GtkMenuItem     *menuitem,
                                         gpointer         user_data)
{
    GtkWidget * dialog;
 
    dialog = gtk_file_chooser_dialog_new (_("please select the directory to save the user phrase library."),
				          NULL,
				          GTK_FILE_CHOOSER_ACTION_SAVE,
				          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				          GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
				          NULL);
					  
    gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
    
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), "/tmp");
    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "scim_pinyin_user_library.txt");

    if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
        gchar    *filename,  *phrase,  *pinyin, *content;
	GtkTreeIter iter;

        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
	
        if (!gtk_tree_model_get_iter_first (tree_model, &iter))
	{
	    gtk_widget_destroy (dialog);
	    return;
	}
	
        content = "#SCIM Pinyin User Phrase Library\n";
	
        do {
            gtk_tree_model_get(tree_model, &iter, PHRASE_COLUMN, &phrase, -1);
            gtk_tree_model_get(tree_model, &iter, PINYIN_COLUMN, &pinyin, -1);
	
	    content = g_strconcat( content, phrase,"    ", pinyin, "\n", NULL);
	
        } while (gtk_tree_model_iter_next (tree_model, &iter));
     
         g_file_set_contents (filename, content, strlen(content), NULL);	 	
    }

    gtk_widget_destroy (dialog);   
}

void
on_button_submit_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
    std::cout << "PinyinHelper: on_button_submit_clicked() \n";
    
    GtkTreeIter iter;
    gchar       *phrase,  *pinyin;
    WideString  w_phrase, w_pinyin;
    std::vector<WideString> phrase_vec;
    std::vector<WideString> pinyin_vec;
    
    if (!gtk_tree_model_get_iter_first (tree_model, &iter))
	return;
	
    do {
        gtk_tree_model_get(tree_model, &iter, PHRASE_COLUMN, &phrase, -1);
        gtk_tree_model_get(tree_model, &iter, PINYIN_COLUMN, &pinyin, -1);

	
	w_phrase = utf8_mbstowcs( phrase );
	w_pinyin = utf8_mbstowcs( pinyin );
	
	phrase_vec.push_back( w_phrase );
	pinyin_vec.push_back( w_pinyin );
	
    } while (gtk_tree_model_iter_next (tree_model, &iter));
    
    if( phrase_vec.size() == 1 || pinyin_vec.size() == 1 )
        return;
    
    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_COMMIT_NEW_PHRASE);	
    send.put_data ( phrase_vec );
    send.put_data ( pinyin_vec );
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
   std::cout << "on_button_submit_clicked() : send : SCIM_PINYIN_TRANS_CMD_COMMIT_NEW_PHRASE\n";

}

void
on_button_add_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
    GtkTreeIter iter;
    GtkTreePath *path; 
    GtkTreeViewColumn* column;    
    
    gtk_widget_show (submit_button);
    gtk_widget_hide (disable_button);
    gtk_widget_hide (add_button);   
    gtk_widget_hide (add_phrase_row);
    gtk_widget_hide (disable_phrase_row);      
    
    g_object_set ( renderer_phrase, "editable", true, NULL);
    g_object_set ( renderer_pinyin, "editable", true, NULL);    
    
    gtk_list_store_clear ( list_store );  
       
    gtk_list_store_append (list_store, &iter);
    gtk_list_store_set (list_store, &iter,
			  NUM_COLUMN, 1,
			  PHRASE_COLUMN, "",
			  PINYIN_COLUMN, "",
			  -1);

    row_count = 1;    
    
    // set cursor to the first phrase cell
    //gtk_tree_model_get_iter_first ( tree_model, &iter);
    path   = gtk_tree_model_get_path  ( tree_model, &iter );    
    column = gtk_tree_view_get_column ( tree_view, 1 );    	
    gtk_tree_view_set_cursor ( tree_view,
                               path,
                               column,
                               TRUE);
			       
    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_GET_MAX_USER_PHRASE_LENGTH);
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
    std::cout << "on_button_add_clicked() - send : SCIM_PINYIN_TRANS_CMD_GET_MAX_USER_PHRASE_LENGTH\n";				  	  
}

void
on_button_disable_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
    GtkTreeIter iter;
    GtkTreeModel *model;
    GtkTreeSelection* selection;
    gchar       *phrase;
    WideString  w_phrase;  
    gboolean    rc;  

    selection = gtk_tree_view_get_selection( tree_view);    

    rc = gtk_tree_selection_get_selected (selection, &model, &iter);
    if( !rc )
        return;
   
    gtk_tree_model_get(model, &iter, PHRASE_COLUMN, &phrase, -1);    
    gtk_tree_model_get(model, &iter, NUM_COLUMN, &list_cur_num, -1);
    
    w_phrase = utf8_mbstowcs( phrase );    
        
    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_DISABLE_PHRASE);
    send.put_data ( w_phrase );
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
    std::cout << "in on_button_disable_clicked(), send : SCIM_PINYIN_TRANS_CMD_DISABLE_PHRASE\n";
}

void
on_phrase_cell_edited			(GtkCellRendererText *renderer,
					 gchar *path,
					 gchar *new_text,
					 gpointer data)
{
/*
    gunichar uni_new_text = g_utf8_get_char ( new_text );    
    PangoScript s = pango_script_for_unichar (uni_new_text);
*/
    if( strlen(new_text) == 0 )
        return;

    if(  g_utf8_strlen (new_text, -1) > max_user_phrase_length ){
	GtkWidget *dialog;
	dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	      				     GTK_DIALOG_MODAL,
					     GTK_MESSAGE_ERROR,
					     GTK_BUTTONS_OK,
					     _("The phrase length exeed the MAX_USER_PHRASE_LENGTH!") );
	gtk_dialog_run( GTK_DIALOG(dialog) );
	gtk_widget_destroy(dialog);    
    
	GtkTreePath *_path = gtk_tree_path_new_from_string (path);
        GtkTreeViewColumn  *column = gtk_tree_view_get_column ( tree_view, 1 );		
        gtk_tree_view_set_cursor ( tree_view,
                                   _path,
                                   column,
                                   TRUE);  	    
	return;    	
    }	
   
    WideString w_new_text;
    
    w_new_text = utf8_mbstowcs( new_text );
  
    gtk_tree_model_get_iter_from_string( tree_model, &cur_iter, path);
 
    gtk_list_store_set(list_store, &cur_iter, 
                       PHRASE_COLUMN, new_text,
                       -1); 
		       
    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_GET_PINYIN);
    send.put_data ( w_new_text );
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
   					
    std::cout << "in on_phrase_cell_edited(), send : SCIM_PINYIN_TRANS_CMD_GET_PINYIN\n";
}

void
on_pinyin_cell_edited			(GtkCellRendererText *renderer,
					 gchar *path,
					 gchar *new_text,
					 gpointer data)
{
    std::cout << "in on_pinyin_cell_edited() " << new_text <<" path="<< path <<"\n" ;
    
    int key_nums;
    gchar  *phrase; 
    String new_text_str;
    String::iterator iter;
    WideString w_phrase;   
    
    
    for(int i = 0; new_text[i]!='\0'; i++){
	
	if( !islower( new_text[i] ) && new_text[i] != ' '){
	    GtkWidget *dialog;
	    dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	    				     GTK_DIALOG_MODAL,
					     GTK_MESSAGE_ERROR,
					     GTK_BUTTONS_OK,
					     _("Invalid pinyin key in the input pinyin sequence.") );
	    gtk_dialog_run( GTK_DIALOG(dialog) );
	    gtk_widget_destroy(dialog);
	    
	    return; 	    					
	}
	
	if( new_text[i] == ' ' && new_text[i+1] == ' ')
	    continue;
	    
	if( new_text[i] == ' ' && new_text[i+1] == '\0')
	    continue;	    
	
	new_text_str.push_back( new_text[i] );	
    } 
    
    iter = new_text_str.begin();
    key_nums = 0;
    while( iter != new_text_str.end() ){
        if( *iter == ' ')
	    key_nums++;
	 iter++;
    }
    if( key_nums == 0 ){
        if( new_text_str.size() > 0 )
            key_nums++;	
    } else {
        key_nums++;    }
        
    gtk_tree_model_get_iter_from_string( tree_model, &cur_iter, path);
    
    gtk_tree_model_get(tree_model, &cur_iter, PHRASE_COLUMN, &phrase, -1);
    
    w_phrase = utf8_mbstowcs( phrase );  
    
    if( w_phrase.size() == 0){
        gtk_list_store_set(list_store, &cur_iter, 
                       	   PINYIN_COLUMN, "",
                           -1);
        return;
    } 
    
    if( w_phrase.size() != key_nums ){
        GtkWidget *dialog;
	dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	        			 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_OK,
					 _("Pinyin keys number is not equal to the phrase charachater number.") );
	    gtk_dialog_run( GTK_DIALOG(dialog) );
	    gtk_widget_destroy(dialog);
	    
	    return;       
     }         
 
    gtk_list_store_set(list_store, &cur_iter, 
                       PINYIN_COLUMN, new_text_str.c_str(),
                       -1); 
      
}

static void
slot_process_imengine_event (const HelperAgent *agent, int ic,
                             const String &uuid, const Transaction &recv)
{
    if (uuid == "05235cfc-43ce-490c-b1b1-c5a2185276ae") {

        int cmd;
        TransactionReader   reader (recv);	
	std::vector<String> pinyin_strings;
	
   	std::vector <String>  	   phrase_pinyin_vec;
    	std::vector <WideString>   phrase_string_vec;
	
        if (reader.get_command (cmd) && cmd == SCIM_TRANS_CMD_REQUEST) {
            while (reader.get_command (cmd)) {
                switch (cmd) {
                    case SCIM_PINYIN_TRANS_CMD_SEND_PINYIN:
                    {
			std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_SEND_PINYIN \n";
			
			if (reader.get_data (pinyin_strings)) {
			
			    gtk_list_store_set(list_store, &cur_iter, 
                       				PINYIN_COLUMN, pinyin_strings[0].c_str(),
                       				-1); 
			} 
			
			guint num;
			gtk_tree_model_get ( tree_model, &cur_iter, 0, &num, -1 );
			
			if ( num == row_count ) {
			    ok_to_add_new_row = true;
			    on_new_row_activate ( NULL,NULL );
			}
			      
                        break;
                    }
                    case SCIM_PINYIN_TRANS_CMD_SEND_PINYIN_FAILED:
                    {
		    	std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_SEND_PINYIN_FAILED \n";
			
    			GtkTreePath 	   *path; 
    			GtkTreeViewColumn  *column;
						
			gtk_list_store_set(list_store, &cur_iter, 
                       			   PHRASE_COLUMN,"",
                       			   -1);	
    			path = gtk_tree_model_get_path ( tree_model, &cur_iter );
			    
    			column = gtk_tree_view_get_column ( tree_view, 1 );    	
			
    			gtk_tree_view_set_cursor ( tree_view,
         			                   path,
     			                           column,
            			                   TRUE); 					   		
                        break;
                    }	
                    case SCIM_PINYIN_TRANS_CMD_SEND_USER_LIB:
                    {
			std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_SEND_USER_LIB \n";
			
			GtkTreeIter iter, focus_iter;
			GtkTreePath *path;
			int i;
			
			/* for multi selection use
    			GtkTreeSelection* selection;
    			selection = gtk_tree_view_get_selection( tree_view); 			
			gtk_tree_selection_set_mode( selection, GTK_SELECTION_MULTIPLE);
			*/
						
			if (reader.get_data (phrase_string_vec) && reader.get_data (phrase_pinyin_vec) ) {

			    gtk_list_store_clear ( list_store );
			    
			    for( i=0 ; i < phrase_string_vec.size(); i++){
			    
			        gtk_list_store_append (list_store, &iter);
				gtk_list_store_set (list_store, &iter,
			  	                    NUM_COLUMN, i+1,
			  			    PHRASE_COLUMN, utf8_wcstombs(phrase_string_vec[i]).c_str(),
			  			    PINYIN_COLUMN, phrase_pinyin_vec[i].c_str(),
			  			    -1);
						    
			        if( list_cur_num == i+1 )
			            focus_iter = iter;
			     }
				
			     // if disable one row, select the next row	
			     if( list_cur_num > 0 ){	  			     
			         if( list_cur_num < i+1 )	
    				     path = gtk_tree_model_get_path ( tree_model, &focus_iter );    
         			 else 
    				     path = gtk_tree_model_get_path ( tree_model, &iter ); 				        
    				 gtk_tree_view_set_cursor ( tree_view, path, NULL, false);				     
				 list_cur_num = -1;
			     }
			} 
                        break;
                    }
                    case SCIM_PINYIN_TRANS_CMD_SAVE_USER_LIB_OK:
                    {
			std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_SAVE_USER_LIB_OK \n";
			
        		GtkWidget *dialog;
			dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	        			 		GTK_DIALOG_MODAL,
					 		GTK_MESSAGE_INFO,
					 		GTK_BUTTONS_OK,
					 		_("Success to save the user phrase library.") );
	    		gtk_dialog_run( GTK_DIALOG(dialog) );
	    		gtk_widget_destroy(dialog);
	
                        break;
                    }
                    case SCIM_PINYIN_TRANS_CMD_SAVE_USER_LIB_FAIL:
                    {
			std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_SAVE_USER_LIB_FAIL \n";
			
        		GtkWidget *dialog;
			dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	        			 		GTK_DIALOG_MODAL,
					 		GTK_MESSAGE_INFO,
					 		GTK_BUTTONS_OK,
					 		_("Failed to save the user phrase library.") );
	    		gtk_dialog_run( GTK_DIALOG(dialog) );
	    		gtk_widget_destroy(dialog);
	
                        break;
                    }		    
                    case SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB_OK:
                    {
			std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB_OK \n";
			
        		GtkWidget *dialog;
			dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	        			 		GTK_DIALOG_MODAL,
					 		GTK_MESSAGE_INFO,
					 		GTK_BUTTONS_OK,
					 		_("Success to import the user phrase library.") );
	    		gtk_dialog_run( GTK_DIALOG(dialog) );
	    		gtk_widget_destroy(dialog);
			
    			Transaction send;
    			send.put_command (SCIM_TRANS_CMD_REQUEST);
    			send.put_command (SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB);
    			helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
   			std::cout << "send : SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB\n";

			break;
                    }
                    case SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB_FAIL:
                    {
			std::cout << "process_imengine_event():get command-  SCIM_PINYIN_TRANS_CMD_IMPORT_USER_LIB_FAIL \n";
			
        		GtkWidget *dialog;
			dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	        			 		GTK_DIALOG_MODAL,
					 		GTK_MESSAGE_INFO,
					 		GTK_BUTTONS_OK,
					 		_("Failed to import the user phrase library.") );
	    		gtk_dialog_run( GTK_DIALOG(dialog) );
	    		gtk_widget_destroy(dialog);
	
                        break;
                    }
                    case SCIM_PINYIN_TRANS_CMD_DISABLE_PHRASE_OK:
                    {
    			Transaction send;
    			send.put_command (SCIM_TRANS_CMD_REQUEST);
    			send.put_command (SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB);
    			helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
	    
   			std::cout << "send : SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB\n" ;
	
                        break;
                    }		    	
                    case SCIM_PINYIN_TRANS_CMD_COMMIT_NEW_PHRASE_OK:
                    {
			std::cout << "process_imengine_event():get command - SCIM_PINYIN_TRANS_CMD_COMMIT_NEW_PHRASE_OK \n";
			
        		GtkWidget *dialog;
			dialog = gtk_message_dialog_new( GTK_WINDOW(BaseWindow), 
	        			 		GTK_DIALOG_MODAL,
					 		GTK_MESSAGE_INFO,
					 		GTK_BUTTONS_OK,
					 		_("Success to add the new phrases into the user phrase library.") );
	    		gtk_dialog_run( GTK_DIALOG(dialog) );
	    		gtk_widget_destroy(dialog);
			
			on_list_user_phrase_library( NULL, NULL);
	
                        break;
                    }	
                    case SCIM_PINYIN_TRANS_CMD_SEND_MAX_USER_PHRASE_LENGTH:
                    {
			std::cout << "process_imengine_event():get command - SCIM_PINYIN_TRANS_CMD_SEND_MAX_USER_PHRASE_LENGTH \n";
		
		        uint32 length;
			if (reader.get_data (length) )	
			    max_user_phrase_length = length;
			
                        break;
                    }		    	    	    		    		    	    	    	    
		    default:
    		    {
        		std::cout << "process_imengine_event():get command-  default\n";
       			break;
    		    }
                }
            }
        }
    }
}

/* create the treeview model*/
static GtkTreeModel * create_model (void)
{
    GtkListStore *store;

    store = gtk_list_store_new (N_COLUMNS,
			        G_TYPE_UINT,
			        G_TYPE_STRING,
			        G_TYPE_STRING);

    list_store = store;
    return GTK_TREE_MODEL (store);
}

/* add columns to the treeview */
static void  add_columns (GtkTreeView *treeview)
{
  GtkTreeViewColumn *column;

  renderer_num = gtk_cell_renderer_text_new ();
  g_object_set ( renderer_num, "xalign", 0.5, NULL);	
  column = gtk_tree_view_column_new_with_attributes (_("ID"),
						     renderer_num,
						     "text",NUM_COLUMN,
						     NULL);
  //g_object_set ( renderer_num, "alignment", 0.5, NULL);
  gtk_tree_view_append_column (treeview, column);

  renderer_phrase = gtk_cell_renderer_text_new ();
  g_object_set ( renderer_phrase, "editable", FALSE, NULL);
  g_signal_connect(renderer_phrase, "edited", (GCallback) on_phrase_cell_edited, NULL);
  column = gtk_tree_view_column_new_with_attributes (_("PHRASE"),
						     renderer_phrase,
						     "text",PHRASE_COLUMN,
						     NULL);
  gtk_tree_view_append_column (treeview, column);

  renderer_pinyin = gtk_cell_renderer_text_new ();
  g_object_set ( renderer_pinyin, "editable", FALSE, NULL);
  g_signal_connect(renderer_pinyin, "edited", (GCallback) on_pinyin_cell_edited, NULL);  
  column = gtk_tree_view_column_new_with_attributes (_("PINYIN"),
						     renderer_pinyin,
						     "text",
						     PINYIN_COLUMN,
						     NULL);
  gtk_tree_view_append_column (treeview, column); 
}

GtkWidget*
create_BaseWindow (void)
{
  GtkWidget *window1;
  GtkWidget *vbox1;
  GtkWidget *menubar1;
  GtkWidget *menuitem1;
  GtkWidget *menuitem1_menu;
  GtkWidget *separatormenuitem1;
  GtkWidget *quit;
  GtkWidget *menuitem2;
  GtkWidget *menuitem4;
  GtkWidget *menuitem2_menu;
  GtkWidget *menuitem4_menu;
  GtkWidget *help_online;
  GtkWidget *new_row;
  GtkWidget *scrolledwindow1;
  GtkWidget *treeview_phrase;
  GtkWidget *hbox1;
  GtkWidget *button_quit;
  GtkWidget *button_submit;
  GtkAccelGroup *accel_group;
  
/* code by hand */  
  GtkTreeModel *model;
  GtkWidget *list_lib_row;  
  GtkWidget *save_lib_row; 
  GtkWidget *import_lib_row; 
  GtkWidget *export_lib_row;
  GtkWidget *button_disable; 
  GtkWidget *button_add; 
/* end */

  accel_group = gtk_accel_group_new ();

  window1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (window1, 400, 300);
  gtk_window_set_title (GTK_WINDOW (window1), _("SCIM Pinyin - Add New Phrase"));

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (window1), vbox1);

  menubar1 = gtk_menu_bar_new ();
  gtk_widget_show (menubar1);
  gtk_box_pack_start (GTK_BOX (vbox1), menubar1, FALSE, FALSE, 0);

  menuitem1 = gtk_menu_item_new_with_mnemonic (_("_Options"));
  gtk_widget_show (menuitem1);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem1);

  menuitem1_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem1), menuitem1_menu);

  separatormenuitem1 = gtk_separator_menu_item_new ();
  gtk_widget_show (separatormenuitem1);
  gtk_container_add (GTK_CONTAINER (menuitem1_menu), separatormenuitem1);
  gtk_widget_set_sensitive (separatormenuitem1, FALSE);

  quit = gtk_image_menu_item_new_from_stock ("gtk-quit", accel_group);
  gtk_widget_show (quit);
  gtk_container_add (GTK_CONTAINER (menuitem1_menu), quit);

  menuitem2 = gtk_menu_item_new_with_mnemonic (_("_Actions"));
  gtk_widget_show (menuitem2);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem2);

/*code by hand */  
  menuitem2_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem2), menuitem2_menu);
  
  list_lib_row = gtk_menu_item_new_with_mnemonic (_("_List User Phrase Library"));
  gtk_widget_show (list_lib_row);
  gtk_container_add (GTK_CONTAINER (menuitem2_menu), list_lib_row);
  
  import_lib_row = gtk_menu_item_new_with_mnemonic (_("_Import User Phrase Library"));
  gtk_widget_show (import_lib_row);
  gtk_container_add (GTK_CONTAINER (menuitem2_menu), import_lib_row);
  
  export_lib_row = gtk_menu_item_new_with_mnemonic (_("_Export User Phrase Library"));
  gtk_widget_show (export_lib_row);
  gtk_container_add (GTK_CONTAINER (menuitem2_menu), export_lib_row);  
  
  add_phrase_row = gtk_menu_item_new_with_mnemonic (_("_Add New Phrases"));
  gtk_widget_show (add_phrase_row);
  gtk_container_add (GTK_CONTAINER (menuitem2_menu), add_phrase_row);
  
  disable_phrase_row = gtk_menu_item_new_with_mnemonic (_("_Delete Phrase"));
  gtk_widget_show (disable_phrase_row);
  gtk_container_add (GTK_CONTAINER (menuitem2_menu), disable_phrase_row);      
/* end */

  menuitem4 = gtk_menu_item_new_with_mnemonic (_("_Help"));
  gtk_widget_show (menuitem4);
  gtk_container_add (GTK_CONTAINER (menubar1), menuitem4);

  menuitem4_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem4), menuitem4_menu);

  help_online = gtk_menu_item_new_with_mnemonic (_("_Help On Line"));
  gtk_widget_show (help_online);
  gtk_container_add (GTK_CONTAINER (menuitem4_menu), help_online);

  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow1);
  gtk_box_pack_start (GTK_BOX (vbox1), scrolledwindow1, TRUE, TRUE, 0);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_SHADOW_IN);

  treeview_phrase = gtk_tree_view_new ();
  gtk_widget_show (treeview_phrase);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), treeview_phrase);
  
/* code by hand */
  tree_view = GTK_TREE_VIEW(treeview_phrase);
  model = create_model ();
  gtk_tree_view_set_model(GTK_TREE_VIEW(treeview_phrase),model);
  g_object_unref (model);
  add_columns (GTK_TREE_VIEW (treeview_phrase));
  tree_model = model;
/* end */

  hbox1 = gtk_hbox_new (TRUE, 0);
  gtk_widget_show (hbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox1, FALSE, FALSE, 0);

  button_quit = gtk_button_new_with_mnemonic (_("Quit"));
  gtk_widget_show (button_quit);
  gtk_box_pack_start (GTK_BOX (hbox1), button_quit, FALSE, FALSE, 0);
  gtk_widget_set_size_request (button_quit, 80, -1);

  button_submit = gtk_button_new_with_mnemonic (_("Submit"));
  //gtk_widget_show (button_submit);
  gtk_box_pack_start (GTK_BOX (hbox1), button_submit, FALSE, FALSE, 0);
  gtk_widget_set_size_request (button_submit, 80, -1);
  submit_button = button_submit;
 
  button_add = gtk_button_new_with_mnemonic (_("Add"));
  gtk_widget_show (button_add);
  gtk_box_pack_start (GTK_BOX (hbox1), button_add, FALSE, FALSE, 0);
  gtk_widget_set_size_request (button_add, 80, -1); 
  add_button = button_add; //qshen add
  
  button_disable = gtk_button_new_with_mnemonic (_("Delete"));
  gtk_widget_show (button_disable);
  gtk_box_pack_start (GTK_BOX (hbox1), button_disable, FALSE, FALSE, 0);
  gtk_widget_set_size_request (button_disable, 80, -1);  
  disable_button = button_disable;//qshen add

  g_signal_connect ((gpointer) quit, "activate",
                    G_CALLBACK (on_quit_activate),
                    NULL);
  g_signal_connect ((gpointer) list_lib_row, "activate",
                    G_CALLBACK (on_list_user_phrase_library),
                    NULL);
  g_signal_connect ((gpointer) import_lib_row, "activate",
                    G_CALLBACK (on_import_user_phrase_library),
                    NULL);
  g_signal_connect ((gpointer) export_lib_row, "activate",
                    G_CALLBACK (on_export_user_phrase_library),
                    NULL);
  g_signal_connect ((gpointer) add_phrase_row, "activate",
                    G_CALLBACK (on_button_add_clicked),
                    NULL);
  g_signal_connect ((gpointer) disable_phrase_row, "activate",
                    G_CALLBACK (on_button_disable_clicked),
                    NULL);		    		    		    		    		    
  g_signal_connect ((gpointer) help_online, "activate",
                    G_CALLBACK (on_help_online_activate),
                    NULL);
  g_signal_connect ((gpointer) button_quit, "clicked",
                    G_CALLBACK (on_button_quit_clicked),
                    NULL);
  g_signal_connect ((gpointer) button_submit, "clicked",
                    G_CALLBACK (on_button_submit_clicked),
                    NULL);
/* code by hand */
  g_signal_connect (G_OBJECT(window1), "destroy",
                    G_CALLBACK (gtk_main_quit),
		    NULL);
  g_signal_connect ((gpointer) button_disable, "clicked",
                    G_CALLBACK (on_button_disable_clicked),
                    NULL);
  g_signal_connect ((gpointer) button_add, "clicked",
                    G_CALLBACK (on_button_add_clicked),
                    NULL);		    		    
/* end */		    
  gtk_window_add_accel_group (GTK_WINDOW (window1), accel_group);


  return window1;
}

static void
run (const String &display, const ConfigPointer &config)
{
    char **argv = new char * [4];
    int    argc = 3;
    
    argv [0] = "smart-pinyin-imengine-helper";
    argv [1] = "--display";
    argv [2] = const_cast<char *> (display.c_str ());
    argv [3] = 0;    
    
    setenv ("DISPLAY", display.c_str (), 1);

    gtk_init (&argc, &argv);

    helper_agent.signal_connect_exit (slot (slot_exit));
    helper_agent.signal_connect_attach_input_context (slot (slot_attach_input_context));
    helper_agent.signal_connect_detach_input_context (slot (slot_detach_input_context));
    helper_agent.signal_connect_update_screen (slot (slot_update_screen));
    helper_agent.signal_connect_update_spot_location (slot (slot_update_spot_location));
    helper_agent.signal_connect_process_imengine_event (slot (slot_process_imengine_event));

    
    BaseWindow = create_BaseWindow ();
    gtk_widget_show (BaseWindow);
    
    // open connection
    int fd = helper_agent.open_connection (helper_info, display);
    GIOChannel *ch = g_io_channel_unix_new (fd);

    if (fd >= 0 && ch) {
        g_io_add_watch (ch, G_IO_IN,
                        helper_agent_input_handler,
                        (gpointer) &helper_agent);
        g_io_add_watch (ch, G_IO_ERR,
                        helper_agent_input_handler,
                        (gpointer) &helper_agent);
        g_io_add_watch (ch, G_IO_HUP,
                        helper_agent_input_handler,
                        (gpointer) &helper_agent);
    }
    
    Transaction send;
    send.put_command (SCIM_TRANS_CMD_REQUEST);
    send.put_command (SCIM_PINYIN_TRANS_CMD_SHOW_USER_LIB);
    helper_agent.send_imengine_event (cur_ic, cur_uuid, send); 
    
    gtk_main ();
    
    // close connection
    helper_agent.close_connection ();
    fd = -1;
}
